#include <stdio.h>
#include <fcntl.h>  /* open()   */
#include <unistd.h> /* read()   */
#include <stdlib.h> /* exit()   */

#include "brpred_core.h"

struct {
  unsigned int pc;
  int type;
  int taken;
} branch[3];

int main(int argc, char **argv) {
 
   int i,fd,branch_type;
   
   unsigned int buffer[1024],max_count=0,pc=0;
   
   unsigned long long cond_branches=0,uncond_branches=0,instructions=0;
   unsigned long long mispredicts=0,likely_branches=0,likely_misses=0;

   /* init counters */
   init_2bit();
   
   fd=open("trace.bpred",O_RDONLY);
   if (fd<0) {
      fprintf(stderr,"Cannot open trace.bpred!\n");
      exit(-1);
   }
   
   while(1) {
      max_count=read(fd,buffer,4096);
      if (max_count<1) break;
      max_count/=8;
      
      instructions+=max_count;
      
      for(i=0;i<max_count;i++) {
         branch_type=get_branch_type(buffer[(i*2)+1]);
	 	 
         if (branch_type==COND_BRANCH) {
	    cond_branches++;
	 }
         else if (branch_type==UNCOND_BRANCH) {
	    uncond_branches++;
	 } 
	 else if (branch_type==LIKELY_BRANCH) { 
	    likely_branches++;
	    cond_branches++;
	 }
	 
	 pc=buffer[i*2];
	 
	 
	 if (branch[1].type==LIKELY_BRANCH) {
	       /* branch was not taken */
	    if (branch[1].pc+8==pc) {
	       likely_misses++;
	       mispredicts++;
	    }			   
	 }
	 
	    /* update counters */
	 if (branch[1].type==COND_BRANCH) {
	    
	    if (branch[1].pc+8==pc) {
	       /* branch was not taken */	       
	       update_2bit(branch[1].pc,0);
	       if (branch[1].taken) {
		  mispredicts++;
	       }
	    }
	    else {
	       /* branch was taken */
	       update_2bit(branch[1].pc,1);
	       if (!branch[1].taken) {
		  mispredicts++;
	       }
	    }
	 }
	    
	 branch[1].type=branch[0].type; branch[1].pc=branch[0].pc;
	 branch[1].taken=branch[0].taken;
	 
	 branch[0].type=branch_type; branch[0].pc=pc;
	 if (branch_type==COND_BRANCH) branch[0].taken=predict_2bit(pc);
	 if (branch_type==LIKELY_BRANCH) branch[0].taken=1;
      }	 
   }
   
   printf("Total instructions: %lld\n",instructions);
   printf("Cond   branches:    %lld\n",cond_branches);
   printf("Uncond branches:    %lld\n",uncond_branches);
   printf("Likely branches:    %lld\n",likely_branches);
   printf("Total misspredicts: %lld\n",mispredicts);
   printf("Likely misspredict: %lld\n",likely_misses);
   printf("Percentage branches:%.2f\n",((double)cond_branches+
					(double)uncond_branches+
					(double)likely_branches)*100.0/
	                                (double)instructions);
   printf("Miss rate:          %.2f\n",(double)mispredicts*100.0/(double)cond_branches);
   printf("Brpred type: 2-bit\n");
   return 0;
}
