#include <stdio.h>
#include <stdlib.h> /* calloc() */

#include "brpred_core.h"

static int *counters;

void init_2bit() {
    int i;
   
    /* init counters */
    counters=calloc(NUM_COUNTERS,sizeof(int));
    for(i=0;i<NUM_COUNTERS;i++) {
       counters[i]=2;
    }
}

int predict_2bit(unsigned int addr) {
   
   register int line;
   
   line=(addr>>COUNT_SHIFT)&COUNT_MASK;

   if (counters[line]>=2) return 1;
   
   return 0;
}

void update_2bit(unsigned int addr,int was_taken) {
   
   register int line;
   
   line=(addr>>COUNT_SHIFT)&COUNT_MASK;
   
   if (was_taken) {
      if (counters[line]<3) counters[line]++;
   }
   else  {
      if (counters[line]>0) counters[line]--;
   }
}


int is_branch_forward(unsigned int insn) {
 
   int direction;
   
   direction=(insn&0xffff)>>15;
   
   if (direction) return 0; /* negative offset */
   else return 1;
   
}

int get_branch_type(unsigned int insn) {
 
   unsigned int opcode,func,rs,rt,nd;
   
   opcode=(insn>>26)&0x3f;
   func=(insn&0x3f);
   rs=(insn>>21)&0x1f;
   rt=(insn>>16)&0x1f;

   
      /* beq */
   if (opcode==0x4) {   
      if ((rs==0)&&(rt==0)) {
	 return UNCOND_BRANCH; /* b */
      
      } else {
         return COND_BRANCH; /* beq */
      }
   }
   
      /* bgezal */
   if ((opcode==0x1) && (rt==0x11)) {
      if (rs==0) return UNCOND_BRANCH; /* bal */
      else return COND_BRANCH; /* begzal */
   }
   
   if ((opcode==0x11) && (rs==0x8)) {
      nd=(insn&(1<<17));  /* "nullify delay slot" AKA "likely" */
      if (nd) return LIKELY_BRANCH; /* bc1fl,bc1tl */
      else return COND_BRANCH; /* bc1f,bc1t */
   }
   
//   if ((opcode==0x12) && (rs==8)) return COND_BRANCH; /* bc2f  */ no c2 ever

   if (opcode==0x14) return LIKELY_BRANCH; /* beql */
   if ((opcode==0x1) && (rt==0x1)) return COND_BRANCH; /* bgez */
   if ((opcode==0x1) && (rt==0x13)) return LIKELY_BRANCH; /* bgezall */
   if ((opcode==0x1) && (rt==0x3)) return LIKELY_BRANCH; /* bgezl */
   if ((opcode==0x7)) return COND_BRANCH; /* bgtz */
   if ((opcode==0x17)) return LIKELY_BRANCH; /* bgtzl */
   if ((opcode==0x6)) return COND_BRANCH; /* blez */
   if ((opcode==0x16)) return LIKELY_BRANCH; /* blezl */   
   if ((opcode==0x1) && (rt==0x0)) return COND_BRANCH; /* bltz */
   if ((opcode==0x1) && (rt==0x10)) return COND_BRANCH; /* bltzal */
   if ((opcode==0x1) && (rt==0x12)) return LIKELY_BRANCH; /* bltzall */
   if ((opcode==0x1) && (rt==0x2)) return LIKELY_BRANCH; /* bltzl */
   if (opcode==0x5) return COND_BRANCH; /* bne */
   if (opcode==0x15) return LIKELY_BRANCH; /* bnel */
   if (opcode==0x2) return UNCOND_BRANCH; /* j */   
   if (opcode==0x3) return UNCOND_BRANCH; /* jal */   
   if ((opcode==0x0) && (func==0x9)) return UNCOND_BRANCH; /* jalr */
   if ((opcode==0x0) && (func==0x8)) return UNCOND_BRANCH; /* jr */
   
   return NOT_A_BRANCH;

}

